package org.netkernel.packge.endpoint;

import java.io.*;
import java.net.URI;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.xml.xda.*;

public class InstallAccessor extends StandardAccessorImpl
{
	public void onSource(INKFRequestContext aHelper) throws Exception
	{
		IXDA page = ((IXDAReadOnly)aHelper.source("res:/installer/pub/formTemplate.xml", IXDAReadOnly.class)).getClone();
		String err=null;
		
		if (!BootUtils.bootedFromJar(aHelper.getKernelContext().getKernel().getConfiguration()))
		{	err = "ERR_NOT_JARBOOT";
		}
		
		IHDSNode param = (IHDSNode)aHelper.source("httpRequest:/params",IHDSNode.class);
		String target=(String)param.getFirstValue("target");
		String proxyHost=(String)param.getFirstValue("proxyHost");
		String proxyPort=(String)param.getFirstValue("proxyPort");
		
		if (target!=null && err==null)
		{	boolean expand=param.getFirstValue("expand")!=null;
			//boolean source=param.getFirstValue("source")!=null;
			boolean source=false;
			
			page.setText("//input[@name='target']/@value", target);
			page.setText("//input[@name='proxyHost']/@value", proxyHost);
			page.setText("//input[@name='proxyPort']/@value", proxyPort);
			
			if (!expand) page.delete("//input[@name='expand']/@checked");
			
			File targetDir = new File(target);
			
			if (!targetDir.exists())
			{	if (!targetDir.mkdirs())
				{	err="ERR_NO_CREATE";
				}
			}
			else if (!targetDir.canWrite())
			{	err="ERR_NO_WRITE";
			}
			
			if (err==null)
			{	if (!targetDir.isDirectory())
				{	err="ERR_NOT_DIR";
				}
			}
			
			if (err==null)
			{	//start install
				String bootJar = BootUtils.getBootedJarName();
				JarFile jarFile = new JarFile(bootJar);
				
				//setup lib directory
				File bootloaderJar=null;
				final String libPrefix="lib/";
				File lib=new File(targetDir,libPrefix);
				lib.mkdir();
				new File(lib, "ext/").mkdir();
				
				List<File> bootFiles = new ArrayList();
			    Enumeration e = jarFile.entries();
			    while (e.hasMoreElements())
			    {
			        JarEntry entry = (JarEntry)e.nextElement();
			        if (entry.isDirectory()) continue;
			        String name = entry.getName();
			        if (!name.startsWith(libPrefix) || !(name.endsWith(".jar") || name.endsWith(".zip"))) continue;
			        
			        InputStream is=jarFile.getInputStream(entry);
			        File bootFile=installJar(is,name,targetDir,false,source);
			        if (bootFile!=null)
			        {	bootFiles.add(bootFile);
			        }
			        if (name.indexOf("core.boot")>=0)
			        {	bootloaderJar=bootFile;
			        }
			    }
				
				//setup modules directory
				final String modulesPrefix="modules/";
				new File(targetDir,modulesPrefix).mkdir();
				List<File> moduleFiles = new ArrayList();
			    e = jarFile.entries();
			    while (e.hasMoreElements())
			    {
			        JarEntry entry = (JarEntry)e.nextElement();
			        if (entry.isDirectory()) continue;
			        String name = entry.getName();
			        if (!name.startsWith(modulesPrefix) || !(name.endsWith(".jar") || name.endsWith(".sjar") || name.endsWith(".zip"))) continue;
			        
			        InputStream is=jarFile.getInputStream(entry);
			        boolean lexpand=expand;
			        if(name.endsWith(".sjar"))
			        {	lexpand=false;
			        }
			        File moduleFile=installJar(is,name,targetDir,lexpand,source);
			        if (moduleFile!=null)
			        {	moduleFiles.add(moduleFile);
			        }
			    }
			    
				//setup etc directory
				File etcDir = new File(targetDir,"etc");
				etcDir.mkdir();
				//setup etc/license directory
				new File(etcDir, "license/").mkdir();
				
				//setup log directory
				new File(targetDir,"log").mkdir();
				
				//setup etc directory
				File appositeDir = new File(targetDir,"etc/apposite");
				appositeDir.mkdir();
				URI targetURI=targetDir.toURI();
				
				//Write apposite database.
				final String appositeDBPrefix="etc/apposite/packagesDB";
				e = jarFile.entries();
			    while (e.hasMoreElements())
			    {
			        JarEntry entry = (JarEntry)e.nextElement();
			        String name = entry.getName();
			        if (!name.startsWith(appositeDBPrefix)) continue;
			        File targetFile=new File(targetDir, name);
			        if (entry.isDirectory())
			        {	targetFile.mkdir();
			        	continue;
			        }
			        InputStream is=jarFile.getInputStream(entry);
			        Utils.pipe(is, new FileOutputStream(targetFile));
			    }
			    if(expand)
			    {	String sql="UPDATE MODULES SET LOCALSRC=REPLACE(LOCALSRC, '.jar', '/' ) WHERE RUNLEVEL>0 AND LOCALSRC NOT LIKE '%.sjar' ;";
			    	File dbpath= new File(targetDir,"etc/apposite/packagesDB");
			    	String cfg="<config>"+
						"<rdbms>"+
							"<jdbcDriver>org.h2.Driver</jdbcDriver>"+
							"<jdbcConnection>jdbc:h2:"+dbpath.getAbsolutePath()+";WRITE_DELAY=0</jdbcConnection>"+
							"<poolSize>1</poolSize>"+
						"</rdbms>"+
					"</config>";
			    	INKFRequest req=aHelper.createRequest("active:sqlUpdate");
			    	req.addArgumentByValue("operand", sql);
			    	req.addArgumentByValue("configuration", cfg);
			    	aHelper.issueRequest(req);
			    }
			    
			    //write bootloader.xml
				File bootloader = new File(etcDir,"bootloader.conf");
				Writer w= new OutputStreamWriter(new FileOutputStream(bootloader));
				w.write("# these file paths relative to the [install] determine classloader to boot kernel\n\n");
				for (File f: bootFiles)
				{	URI fu=targetURI.relativize(f.toURI());
					w.write(fu.toString());
					w.write('\n');
				}
			    w.close();
			    
			    String s;
			    //copy stem-system.conf
			    if(expand)
			    {	s = readEntryAsString(jarFile, "etc/stem-system-expanded.conf");
			    }
			    else
			    {	s = readEntryAsString(jarFile, "etc/stem-system-jarred.conf");
			    }
			    writeStringToFile(s, new File(etcDir,"stem-system.conf"));
			    
			    //write modules.xml
			    if(expand)
			    {	s = readEntryAsString(jarFile, "etc/modules-expanded.xml");
			    }
			    else
			    {	s = readEntryAsString(jarFile, "etc/modules-jarred.xml");
			    }
			    writeStringToFile(s, new File(etcDir,"modules.xml"));
			    
			    
			    //copy kernel.properties
			    File expandedDir = new File(targetDir,"lib/expanded/");
			    expandedDir.mkdir();
			    s = readEntryAsString(jarFile, "etc/kernel.properties");
			    String path=expandedDir.getAbsolutePath();
			    path=path.replaceAll("\\\\","\\\\\\\\\\\\\\\\"); //escape backslash for both regex and property files
			    s=s.replaceAll("netkernel.layer0.expandDir=", "netkernel.layer0.expandDir="+path);
			    if(proxyHost.length()>0)
			    {	s=s.replaceAll("%HTTPPROXYHOST%", proxyHost);
			    }
			    if(proxyHost.length()>0)
			    {	s=s.replaceAll("%HTTPPROXYPORT%", proxyPort);
			    }
			    writeStringToFile(s, new File(etcDir,"kernel.properties"));
			    
			    //Write log config
			    s = readEntryAsString(jarFile, "etc/KernelLogConfig.xml");
			    writeStringToFile(s, new File(etcDir,"KernelLogConfig.xml"));
			    
			    //Write log-level file out
			    s = readEntryAsString(jarFile, "etc/logLevels.xml");
			    writeStringToFile(s, new File(etcDir,"logLevels.xml"));

			    
				File binDir = new File(targetDir,"bin");
				binDir.mkdir();
				
				//Write bootjar.cnf
				s = readEntryAsString(jarFile, "bin/bootjar.cnf");
				s=s.replaceAll("%BOOTLOADER%", bootloaderJar.getName());
				File scriptTarget=new File(binDir,"bootjar.cnf");
				writeStringToFile(s, scriptTarget);
				
				//Write jvmsettings.cnf
				s = readEntryAsString(jarFile, "bin/jvmsettings.cnf");
				scriptTarget=new File(binDir,"jvmsettings.cnf");
				writeStringToFile(s, scriptTarget);
				
				if (File.separatorChar=='/')
				{	// bin/netkernel.sh
				    path=targetDir.getAbsolutePath();
					s = readEntryAsString(jarFile, "bin/netkernel.sh");
					s=s.replaceAll("%INSTALL%", path);
					scriptTarget=new File(binDir,"netkernel.sh");
					writeStringToFile(s, scriptTarget);
					
					Runtime.getRuntime().exec(new String[] {"chmod","u+x",scriptTarget.getAbsolutePath()});
				}
				else
				{	// bin/netkernel.bat
					s = readEntryAsString(jarFile, "bin/netkernel.bat");
					path=targetDir.getAbsolutePath();
					path=path.replaceAll("\\\\","\\\\\\\\");
					s=s.replaceAll("%INSTALL%", path);
					writeStringToFile(s, new File(binDir,"netkernel.bat"));
					
					// bin/netkernel-debug.bat
					s = readEntryAsString(jarFile, "bin/netkernel-debug.bat");
					s=s.replaceAll("%INSTALL%", path);
					writeStringToFile(s, new File(binDir,"netkernel-debug.bat"));
				}
			    
			    page = ((IXDAReadOnly)aHelper.source("res:/installer/pub/installedTemplate.xml", IXDAReadOnly.class)).getClone();
			    page.setText("//span[@id='target']", targetDir.getAbsolutePath());
			}			
		}

		
		if (err!=null)
		{	String msg = aHelper.formatMessage(err, null);
			page.setText("//div[@id='msg']", msg);
			page.setText("//div[@id='msg']/@class","msg_error");
		}		
		
		INKFResponse resp=aHelper.createResponseFrom(page);
		resp.setMimeType("text/xml");
		resp.setHeader("WrappedControlPanel","Install to Disk");
	}
	
	private static String readEntryAsString(JarFile aJarFile, String aName) throws IOException
	{	InputStream is = aJarFile.getInputStream(aJarFile.getEntry(aName));
	    InputStreamReader sr = new InputStreamReader(is);
	    StringWriter sw = new StringWriter(1024);
	    int c;
	    char[] cbuf = new char[256];
	    while ((c=sr.read(cbuf))>0)
	    {	sw.write(cbuf,0,c);
	    }
	    return sw.toString();
	}
	
	private static void writeStringToFile(String aString, File aFile) throws IOException
	{	ByteArrayInputStream bais = new ByteArrayInputStream(aString.getBytes());
	    FileOutputStream fos = new FileOutputStream(aFile);
	    Utils.pipe(bais, fos);
	}
	
	private static File installJar(InputStream aSource, String aName, File aTargetDir, boolean aExpand, boolean aIncludeSource) throws Exception
	{
		File result=null;
		final String SRC_TAG = "-src";
		boolean isSource=aName.contains(SRC_TAG);
		if (isSource && !aIncludeSource) return null;
		
		File targetFile = new File(aTargetDir,aName);
		Utils.pipe(aSource, new FileOutputStream(targetFile));
		
		if (aExpand)
		{	File expandDir;
			if (isSource)
			{	int i=aName.indexOf(SRC_TAG);
				String name=aName.substring(0, i)+aName.substring(i+SRC_TAG.length(), aName.length()-4);
				expandDir = new File(aTargetDir,name);
			}
			else
			{	expandDir = new File(aTargetDir,aName.substring(0, aName.length()-4));
			}
		
			expandDir.mkdir();
			JarFile jarFile = new JarFile(targetFile);
			Enumeration e = jarFile.entries();
		    while (e.hasMoreElements())
		    {  JarEntry entry = (JarEntry)e.nextElement();
		        if (entry.isDirectory()) continue;
		        String name = entry.getName(); 
		        if (name.startsWith("META-INF")) continue;
		        InputStream is=jarFile.getInputStream(entry);
		        File dest = new File(expandDir,name);
		        dest.getParentFile().mkdirs();
		        Utils.pipe(is, new FileOutputStream(dest));
		    }
		    jarFile.close();
			targetFile.delete();
			result=expandDir;
		}
		else
		{	result=targetFile;
		}
		if (isSource) result=null;
		return result;
	}
	
	private class RestartThread extends Thread
	{
		public RestartThread()
		{
		}
	}
}
